/***********************************************************************
DP utilities
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "useful.h"
#include "dynamics.h"
#include "dp.h"

/**********************************************************************/
/**********************************************************************/

/* SPEEDUP STUFF */
/* 1/(increment for each state) */
static float one_over_delta[N_STATE_DIMENSIONS];
/* Speedup: table lookup of index_to_state() */
#define MAX_STATE_RESOLUTION 1002
static float index_to_state_array[MAX_STATE_RESOLUTION][N_STATE_DIMENSIONS];
static int neighbor_index[N_NEIGHBORS]; /* value function index of neighbors */
static int neighbor_indices[N_NEIGHBORS][N_STATE_DIMENSIONS];

/* bounds used in determining when to apply LQR control */
static float lqr_low[N_STATE_DIMENSIONS] = { M_PI, 0.0 };
static float lqr_high[N_STATE_DIMENSIONS] = { 0.0, 0.0 };

extern float model_next_states[N_TOTAL_STATES][N_STATE_DIMENSIONS][N_ACTION_DIMENSIONS+1];

/************************************************************************/
/* Calculate a distance^2 between two states */

float state_distance2( float *s1, float *s2 )
{
  float result = 0;
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    result += ( s1[i] - s2[i] )*( s1[i] - s2[i] );
  
  return result;
}

/************************************************************************/
/* Calculate a normalized distance^2 between two states */

float normalized_state_distance2( float *s1, float *s2 )
{
  float diff;
  float result = 0;
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      diff = ( s1[i] - s2[i] )*one_over_delta[i];
      result += diff*diff;
    }
  return result;
}

/************************************************************************/
/* conversion from cell index to state value in one dimension */
/* This should be a table lookup */

float index_to_state( int dimension, int index )
{
  return min_states[dimension] 
	  + (max_states[dimension] - min_states[dimension])*((float)index)
	  /((float)(n_states[dimension]) - 1.0);
}

/************************************************************************/
/* conversion from a set of indices of a cell to the corresponding state */

int indices_to_state( int *indices, float *state )
{
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    state[i] = index_to_state( i, indices[i] );
  return 1;
}

/************************************************************************/
/* conversion from state value to cell index in one dimension */

int state_to_index( int dimension, float state )
{
  return floorf((state - min_states[dimension])*one_over_delta[dimension]);
}

/************************************************************************/
/* conversion from state to a set of indices of a cell */

int state_to_indices( float *state, int *indices )
{
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    indices[i] = state_to_index( i, state[i] );
  return 1;
}

/************************************************************************/
/* conversion from index to action value in one dimension
 when using discretized actions */
/* This should be a table lookup */

float index_to_action( int dimension, int index, int *n_actions )
{
  return min_actions[dimension] 
    + (max_actions[dimension] - min_actions[dimension])*((float)index)
    /((float)(n_actions[dimension]) - 1.0);
}

/************************************************************************/
/* conversion from indices to action when using discretized actions */

int indices_to_action( int *indices, float *action, int *n_actions )
{
  int i;

  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    action[i] = index_to_action( i, indices[i], n_actions );
  return 1;
}

/************************************************************************/
/* convert cell indices to a one dimensional index */

int value_function_index( int *indices )
{
  return indices[0]*n_states[1] + indices[1];
}

/************************************************************************/
/* convert one dimensionsional cell index to state */

int value_function_index_to_state( int index, float *state )
{
  int indices[N_STATE_DIMENSIONS];

  indices[0] = index/n_states[1];
  indices[1] = index - indices[0]*n_states[1];
  indices_to_state( indices, state );
  return TRUE;
}

/************************************************************************/
/* Copy cell indices */

int copy_state_indices( int *indices, int *result )
{
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    result[i] = indices[i];
  return 1;
}

/************************************************************************/
/* Add a pair of cell indices */

int sum_state_indices( int *indices1, int *indices2, int *result )
{
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    result[i] = indices1[i] + indices2[i];
  return 1;
}

/************************************************************************/
/* write array to output file */

int write_array_to_file( char *name, float *array, int size )
{
  int ip, iv, i1d;
  FILE *output;
  int n_written;

  output = fopen( name, "w" );
  if ( output == NULL )
    {
      fprintf( stderr, "Couldn't open file %s.\n", name );
      exit( -1 );
    }
  n_written = fwrite( (void *) array, sizeof ( float ), size, output );
  fclose( output );

  if ( n_written < size )
    {
      fprintf( stderr, "Write size error %d < %d\n", n_written, size );
      exit( -1 );
    }

  return 0;
}

/************************************************************************/
/* read an array from a file */

int read_array_from_file( char *name, float *array, int size )
{
  int ip, iv, i1d;
  int n_read;
  FILE *input;

  input = fopen( name, "r" );
  if ( input == NULL )
    {
      fprintf( stderr, "Couldn't open file %s.\n", name );
      exit( -1 );
    }

  n_read = fread( (void *) array, sizeof ( float ), size, input );
  fclose( input );

  if ( n_read < size )
    {
      fprintf( stderr, "Read size error %d < %d\n", n_read, size );
      exit( -1 );
    }

  return 0;
}

/************************************************************************/
/* Initialization for these routines */

int init_dp()
{
  int indices[N_STATE_DIMENSIONS];
  int indices_tmp[N_STATE_DIMENSIONS];
  float state_tmp[N_STATE_DIMENSIONS];
  int index;
  int i, j;

  /* Speedup stuff */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      one_over_delta[i] = ((float)(n_states[i]) - 1.0)
	/(max_states[i] - min_states[i]);
    }

  /* Manage LQR boundary */
  state_to_indices( the_desired_state, indices_tmp );
  indices_to_state( indices_tmp, state_tmp );
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      index = indices_tmp[i];
      if ( ( the_desired_state[i] - state_tmp[i] ) 
	   < LQR_PARAM/one_over_delta[i] )
	{
	  state_tmp[i] -= 1/one_over_delta[i];
	  --index;
	}
      if ( lqr_low[i] > state_tmp[i] )
	{
	  lqr_low[i] = state_tmp[i] - LQR_BORDER;
	  printf( "lqr_low[%d] %d -> %g\n", i, index, lqr_low[i] );
	}
      else
	printf( "lqr_low[%d] %g\n", i, lqr_low[i] );
      indices_tmp[i] += 1;
    }
  indices_to_state( indices_tmp, state_tmp );
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      index = indices_tmp[i];
      if ( ( state_tmp[i] - the_desired_state[i] ) 
	   < LQR_PARAM/one_over_delta[i] )
	{
	  state_tmp[i] += 1/one_over_delta[i];
	  ++index;
	}
      if ( lqr_high[i] < state_tmp[i] )
	{
	  lqr_high[i] = state_tmp[i] + LQR_BORDER;
	  printf( "lqr_high[%d] %d -> %g\n", i, index, lqr_high[i] );
	}
      else
	printf( "lqr_high[%d] %g\n", i, lqr_high[i] );
    }

  /* Speedup stuff */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      for( j = 0; j < n_states[i]; j++ )
	index_to_state_array[j][i] = index_to_state( i, j );
    }

  /* Speedup stuff */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    indices[i] = 0;

  copy_state_indices( indices, indices_tmp );

  copy_state_indices( indices_tmp, &(neighbor_indices[0][0]) );
  neighbor_index[0] = value_function_index( indices_tmp );
  indices_tmp[1] += 1;
  copy_state_indices( indices_tmp, &(neighbor_indices[1][0]) );
  neighbor_index[1] = value_function_index( indices_tmp );
  indices_tmp[1] -= 1;

  indices_tmp[0] += 1;
  copy_state_indices( indices_tmp, &(neighbor_indices[2][0]) );
  neighbor_index[2] = value_function_index( indices_tmp );
  indices_tmp[1] += 1;
  copy_state_indices( indices_tmp, &(neighbor_indices[3][0]) );
  neighbor_index[3] = value_function_index( indices_tmp );
  indices_tmp[1] -= 1;
  indices_tmp[0] -= 1;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      if ( indices[i] != indices_tmp[i] )
	{
	  fprintf( stderr, "You screwed up indices_tmp. %d %d %d\n",
		   i, indices[i], indices_tmp[i] );
	  exit( -1 );
	}
    }

  return 1;
}

/************************************************************************/
/*
predict_next_state() is where we either used the precomputed model
or a more general model, depending on whether the state corresponds
to a cell location or not.
*/
/* This would be faster on repeated calls with the same state
   if the indices arg was a pointer to the appropropriate cell */

int predict_next_state( float *state, int index, float *action, 
			float *next_state )
{
  int i, j;

  /* If we are given state, then integrate using model from dynamics.c */
  if ( state != NULL )
    {
      set_state( 0.0, state );
      integrate_one_step( action, next_state );
      return TRUE;
    }

  /* We are on a cell, use stored model in model_next_states */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      next_state[i] = model_next_states[index][i][0];
      for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	next_state[i] += model_next_states[index][i][j+1]*action[j];
    }
  return TRUE;
}

/************************************************************************/
/* Are we in the region where we will use the LQR controller? */

int in_lqr_box( float *state )
{
  int i;

  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      if ( state[i] < lqr_low[i] || state[i] > lqr_high[i] )
	return FALSE;
    }
  return TRUE;
}

/************************************************************************/
/* Value function defined by LQR controller */

float lqr_value( float *state )
{
  int i, j;
  float temp[N_STATE_DIMENSIONS];
  float value;
  float lqr_s[N_STATE_DIMENSIONS][N_STATE_DIMENSIONS] =
    { { 13.2088, 3.2767 },
      { 3.2767, 0.8136 } };

  /* Compute value */
  for ( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      temp[i] = 0;
      for( j = 0; j < N_STATE_DIMENSIONS; j++ )
	temp[i] += lqr_s[i][j]*( state[j] - the_desired_state[j] );
    }
  value = 0;
  for( j = 0; j < N_STATE_DIMENSIONS; j++ )
    value += temp[j]*( state[j] - the_desired_state[j] );

  return value;
}

/************************************************************************/
/* Policy defined by LQR controller */

int lqr_action( float *state, float *action )
{
  int i, j;
  float lqr_k[N_ACTION_DIMENSIONS][N_STATE_DIMENSIONS] =
    { { 9.6438, 2.3947 } };

  for ( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    {
      action[i] = 0;
      for( j = 0; j < N_STATE_DIMENSIONS; j++ )
	action[i] -= lqr_k[i][j]*( state[j] - the_desired_state[j] );
    }

  return TRUE;
}

/************************************************************************/
/* Are we in the LQR region? If so, use LQR value function and policy */

int lqr_covers_it( int *indices, float *state, float *v_best, 
		   float *best_action,
		   float *best_next_state )
{
  float value;
  int index;
  float state_backup[N_STATE_DIMENSIONS];

  if ( indices != NULL )
    {
      state = state_backup;
      indices_to_state( indices, state );
    }

  value = lqr_value( state );
  /* Use combination of ellipsoid and a box containing goal as 
     shape test to apply LQR controller */
  if ( value > LQR_VALUE_THRESHOLD && !in_lqr_box( state ) )
    return FALSE;

  if ( v_best != NULL )
    *v_best = value;

  if ( best_action == NULL )
    return FALSE;

  /* compute action */
  lqr_action( state, best_action );

  /* compute next state */
  if ( best_next_state != NULL )
    {
      if ( indices != NULL )
	{
	  index = value_function_index( indices );
	  predict_next_state( NULL, index, best_action, best_next_state );
	}
      else
	{
	  predict_next_state( state, -1, best_action, best_next_state );
	}
    }

  return TRUE;
}

/************************************************************************/
/************************************************************************/
/* lookup_value looks up the value defined by the value function using
multilinear interpolation over a rectangular grid. */

float lookup_value( float *state, float *v_array )
{
  int i;
  int indices[N_STATE_DIMENSIONS];
  int indices_tmp[N_STATE_DIMENSIONS];
  int base_index = 0;
  float neighbor_state[N_STATE_DIMENSIONS];
  int bad_values;
  float w[N_STATE_DIMENSIONS][2];
  float value;
  float sum_weights = 0.0;
  float sum_values = 0.0;
  float weight = 0;
  float weights[N_NEIGHBORS]; /* for debugging */
  float distance2 = 0;

  // printf( "\nlookup_value: %lg %lg\n", state[0], state[1] );

  /* Note conversion to integer */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      indices[i] = state_to_index( i, state[i] );
      /* Handle situations with states out of bounds. */
      if ( indices[i] < 0 )
	return INFINITE_COST;
      if( indices[i] >= n_states[i] - 1 )
	return INFINITE_COST;
    }
  // printf( "indices OK: %d %d\n", indices[0], indices[1] );

  base_index = value_function_index( indices );

  bad_values = 0;
  for( i = 0; i < N_NEIGHBORS; i++ )
    {
      if ( v_array[base_index + neighbor_index[i]] >= COST_CUTOFF )
	bad_values++;
    }
  // printf( "bad_values: %d\n", bad_values );

  if ( bad_values > 0 )
    {
      for( i = 0; i < N_NEIGHBORS; i++ )
	{
	  // We have a bad value. Compute weighted average.
	  // could get rid of neighbor indices by having lookup based on i
	  // and precomputed table.
	  // just make a table of distances by summing components.
	  if ( v_array[base_index + neighbor_index[i]] < COST_CUTOFF )
	    {
	      sum_state_indices( indices, &(neighbor_indices[i][0]),
				 indices_tmp );
	      indices_to_state( indices_tmp, neighbor_state );
	      weight = expf( WEIGHTED_AVE_SCALE
			     *normalized_state_distance2( state, 
							  neighbor_state ) );
	      weights[i] = weight;
	      sum_values += v_array[base_index + neighbor_index[i]]*weight;
	      sum_weights += weight;
	    }
	  else
	    weights[i] = 0;
	}
      if ( sum_weights != 0.0 )
	value = sum_values/sum_weights;
      else
	value = INFINITE_COST;

      if ( value < 0 )
	{
	  printf( "lookup_value weighted average: Negative V: %lg\n", value );
	  printf( "state %d %d: %lg %lg\n", indices[0], indices[1], 
		  state[0], state[1] );
	  printf( "weights: %lg %lg %lg %lg\n", weights[0], weights[1],
		  weights[2], weights[3] );
	  printf( "%lg %lg %lg\n", sum_values, sum_weights, value );
	  printf( "Press return to continue.\n" );
	  getchar();
	}
      return value;
    }

  /* Compute weights. */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      w[i][1] = (state[i] - index_to_state_array[indices[i]][i])
	*one_over_delta[i];
      w[i][0] = 1.0 - w[i][1];
    }

  /* All values available, do multilinear interpolation */
  value = w[0][0]*w[1][0]*v_array[base_index + neighbor_index[0]]
    + w[0][0]*w[1][1]*v_array[base_index + neighbor_index[1]]
    + w[0][1]*w[1][0]*v_array[base_index + neighbor_index[2]]
    + w[0][1]*w[1][1]*v_array[base_index + neighbor_index[3]];

  // printf( "%lg %lg %lg\n", w[0][0], w[1][0], value );

  if ( value < 0 )
    {
      printf( "lookup_value: Negative V: %lg\n", value );
      printf( "state %d %d: %lg %lg\n", indices[0], indices[1], 
	      state[0], state[1] );
      printf( "weights: %lg %lg %lg %lg\n", w[0][0], w[0][1],
	      w[1][0], w[1][1] );
      printf( "Press return to continue.\n" );
      getchar();
    }

  return value;
}

/************************************************************************/
/* lookup_action looks up the action defined by the policy using
multilinear interpolation over a rectangular grid. */

int lookup_action( float *state,
		   float a_array[N_TOTAL_STATES][N_ACTION_DIMENSIONS],
		   float *v_array,
		   float *action )
{
  int i, j;
  int indices[N_STATE_DIMENSIONS];
  int indices_tmp[N_STATE_DIMENSIONS];
  int base_index = 0;
  float neighbor_state[N_STATE_DIMENSIONS];
  int bad_values;
  float w[N_STATE_DIMENSIONS][2];
  float value;
  float sum_weights = 0.0;
  float sum_values = 0.0;
  float weight = 0;
  float weights[N_NEIGHBORS]; /* for debugging */
  float distance2 = 0;

  /*
  printf( "\nlookup_action: %g %g %g %g\n",
	  state[0], state[1], state[2], state[3] );
  */

  /* Note conversion to integer */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      indices[i] = state_to_index( i, state[i] );
      /* Handle situations with states out of bounds. */
      if ( indices[i] < 0 )
	{
	  for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	    action[j] = BAD_U;
	  return FALSE;
	}
      if ( indices[i] >= n_states[i] - 1 )
	{
	  for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	    action[j] = BAD_U;
	  return FALSE;
	}
    }
  /*
  printf( "indices OK: %d %d %d %d\n", indices[0], indices[1],
	  indices[2], indices[3] );
  */

  base_index = value_function_index( indices );

  bad_values = 0;
  for( i = 0; i < N_NEIGHBORS; i++ )
    {
      if ( v_array[base_index + neighbor_index[i]] >= COST_CUTOFF )
	bad_values++;
    }
  /*
  printf( "bad_values: %d\n", bad_values );
  */

  if ( bad_values > 0 )
    {
      for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	{
	  for( i = 0; i < N_NEIGHBORS; i++ )
	    {
	  // We have a bad value. Compute weighted average.
	  // could get rid of neighbor indices by having lookup based on i
	  // and precomputed table.
	  // just make a table of distances by summing components.
	      if ( v_array[base_index + neighbor_index[i]] < COST_CUTOFF )
		{
		  sum_state_indices( indices, &(neighbor_indices[i][0]),
				     indices_tmp );
		  indices_to_state( indices_tmp, neighbor_state );
		  weight = expf( WEIGHTED_AVE_SCALE*normalized_state_distance2(
					     state, neighbor_state ) );
		  weights[i] = weight;
		  sum_values += 
		    a_array[base_index + neighbor_index[i]][j]*weight;
		  sum_weights += weight;
		}
	      else
		weights[i] = 0;
	    }
	  if ( sum_weights != 0.0 )
	    action[j] = sum_values/sum_weights;
	  else
	    action[j] = BAD_U;
	}
      return TRUE;
    }

  /* Compute weights. */
  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
    {
      w[i][1] = (state[i] - index_to_state_array[indices[i]][i])
	*one_over_delta[i];
      w[i][0] = 1.0 - w[i][1];
    }

  /* All values available, do multilinear interpolation */
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    {
      action[i] = w[0][0]*w[1][0]*a_array[base_index + neighbor_index[0]][i]
	+ w[0][0]*w[1][1]*a_array[base_index + neighbor_index[1]][i]
	+ w[0][1]*w[1][0]*a_array[base_index + neighbor_index[2]][i]
	+ w[0][1]*w[1][1]*a_array[base_index + neighbor_index[3]][i];
    }

  return TRUE;
}

/************************************************************************/
/* update_action does a value function backup using the current policy:
V(x) = c(x,u) + V(f(x))
This is useful to make sure a cell is up to date with respect to other
value function updates that may have occurred.
*/

int update_action( int *indices, float *v_best, float *action,
		   float *next_state, float *value_array )
{
  float state[N_STATE_DIMENSIONS];
  float value;

  indices_to_state( indices, state );
  if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
    value = lookup_value( next_state, value_array );

  value = one_step_cost( state, action ) + DISCOUNT*value;
  *v_best = value;

  return TRUE;
}

/************************************************************************/

static int gradient_debug = 0;
#define GRAD_DELTA_DIVIDER 1000.0
#define GRAD_INITIAL_STEP_SIZE 0.001

/************************************************************************/
/* gradient_descent does a simple gradient descent search to improve the
action at a cell, given the current value function.
best_action must not be NULL, it is the
starting point of the search. best_next_state can be NULL */

int gradient_descent( int *indices, float *state, 
		      float *v_best, float *best_action,
		      float *best_next_state, float *value_array )
{
  int i, j;
  float action[N_ACTION_DIMENSIONS];
  float next_state[N_STATE_DIMENSIONS];
  float delta[N_ACTION_DIMENSIONS];
  float gradient[N_ACTION_DIMENSIONS];
  float v0;
  float a0[N_ACTION_DIMENSIONS];
  float value;
  float gradient_length2 = 0;
  float scale = 0;
  float step_size = GRAD_INITIAL_STEP_SIZE;
  int index;
  float state_backup[N_STATE_DIMENSIONS];

  /* Give up if initial value bad */
  if ( *v_best > COST_CUTOFF )
    return 0;

  if ( indices != NULL )
    { /* Indices is valid, use it to generate state. */
      state = state_backup;
      indices_to_state( indices, state );
      index = value_function_index( indices );
    }
  else
    { /* state is valid, indices is not */
      if ( state == NULL )
	{
	  fprintf( stderr, "NULL state in gradient_descent()\n" );
	  exit( -1 );
	}
    }

  if ( gradient_debug )
    printf( "gradient: %g %g %g %g; %g %g %g\n", *v_best, state[0], state[1],
	    best_action[0], min_actions[0], max_actions[0], delta[0] );

  /* Stash nominal value and action */
  v0 = *v_best;
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    {
      a0[i] = best_action[i];
      action[i] = best_action[i];
    }

  /* Calculate the gradient */
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    {
      delta[i] = (max_actions[i] - min_actions[i])/GRAD_DELTA_DIVIDER;
      action[i] += delta[i];
      if ( indices != NULL )
	predict_next_state( NULL, index, action, next_state );
      else
	predict_next_state( state, index, action, next_state );
      if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
	value = lookup_value( next_state, value_array );
      value = one_step_cost( state, action ) + DISCOUNT*value;
      gradient[i] = (value - v0)/delta[i];
      if ( value > COST_CUTOFF )
	{ /* Try other direction. */
	  action[i] = a0[i] - delta[i];
	  if ( indices != NULL )
	    predict_next_state( NULL, index, action, next_state );
	  else
	    predict_next_state( state, index, action, next_state );
	  if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
	    value = lookup_value( next_state, value_array );
	  value = one_step_cost( state, action ) + DISCOUNT*value;
	  gradient[i] = -(value - v0)/delta[i];
	}
      if ( value > COST_CUTOFF )
	{
	  /* Could compute gradient */
	  return 0;
	}
      if( value < *v_best )
	{
	  *v_best = value;
	  for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	    best_action[j] = action[j];
	  for( j = 0; j < N_STATE_DIMENSIONS; j++ )
	    best_next_state[j] = next_state[j];
	}
      action[i] = a0[i];
    }
  if ( gradient_debug )
    printf( "grad 2: %g %g; %g %g; %g %g\n", a0[0], v0, action[0], value,
	    gradient[0], delta[0] );

  /* Normalize the gradient */
  gradient_length2 = 0;
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    gradient_length2 += gradient[i]*gradient[i];
  if ( gradient_length2 < 1e-6 )
    return 0;
  scale = 1.0/sqrt( gradient_length2 );
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    gradient[i] *= scale;
  if ( gradient_debug )
    printf( "scale: %g %g %g\n", gradient_length2, scale, gradient[0] );

  /* Now search. */
  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    a0[i] = best_action[i];
  for( ; ; )
    {
      for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
	action[i] = a0[i] - step_size*gradient[i];
      if ( action[i] > max_actions[i] )
	break;
      if ( action[i] < min_actions[i] )
	break;
      if ( indices != NULL )
	predict_next_state( NULL, index, action, next_state );
      else
	predict_next_state( state, index, action, next_state );
      if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
	value = lookup_value( next_state, value_array );
      value = one_step_cost( state, action ) + DISCOUNT*value;
      if ( gradient_debug )
	printf( "Trying %g %g %20.10f\n", step_size, value, action[0] );
      if( value < *v_best )
	{
	  *v_best = value;
	  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
	    {
	      best_action[i] = action[i];
	      a0[i] = action[i];
	    }
	  if ( best_next_state != NULL )
	    for( i = 0; i < N_STATE_DIMENSIONS; i++ )
	      best_next_state[i] = next_state[i];
	  step_size *= 2.1;
	}
      else
	{
	  step_size *= 0.1;
	  if ( step_size < 1e-4 ) /* Should be scaled to problem */
	    break;
	}
    }
  if ( gradient_debug )
    {
      printf( "Press return to continue.\n" );
      getchar();
    }
  return 1;
}

/************************************************************************/
/*
random_search generates a single random action and evaluates it.
*/

int random_search( int *indices, float *v_best, float *best_action,
		   float *best_next_state, float *value_array )
{
  int i, j;
  float action[N_ACTION_DIMENSIONS];
  float state[N_STATE_DIMENSIONS];
  float next_state[N_STATE_DIMENSIONS];
  float value;
  int index;

  indices_to_state( indices, state );
  index = value_function_index( indices );

  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
    action[i] = 
      ((float) random()/ (float) RAND_MAX)
      *(max_actions[i] - min_actions[i])
      + min_actions[i];

  predict_next_state( NULL, index, action, next_state );

  if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
    value = lookup_value( next_state, value_array );

  value = one_step_cost( state, action ) + DISCOUNT*value;

  if( value < *v_best )
    {
      *v_best = value;
      for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
	best_action[i] = action[i];
      for( i = 0; i < N_STATE_DIMENSIONS; i++ )
	best_next_state[i] = next_state[i];
    }
  return TRUE;
}

/************************************************************************/
/* grid_search performs a search of a rectangular grid of actions. */

int grid_search( int *indices, float *v_best, float *best_action,
		 float *best_next_state, float *value_array, int *n_actions )
{
  int i, j;
  int action_indices[N_ACTION_DIMENSIONS];
  float action[N_ACTION_DIMENSIONS];
  float state[N_STATE_DIMENSIONS];
  float next_state[N_STATE_DIMENSIONS];
  float value;
  int index;

  *v_best = INFINITE_COST;
  indices_to_state( indices, state );
  index = value_function_index( indices );

  for( action_indices[0] = 0; action_indices[0] < n_actions[0];
       (action_indices[0])++ )
    {
      indices_to_action( action_indices, action, n_actions );
      predict_next_state( NULL, index, action, next_state );

      if ( !lqr_covers_it( NULL, next_state, &value, NULL, NULL ) )
	value = lookup_value( next_state, value_array );
      
      value = one_step_cost( state, action ) + DISCOUNT*value;

      if( value < *v_best )
	{
	  *v_best = value;
	  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
	    best_action[i] = action[i];
	  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
	    best_next_state[i] = next_state[i];
	}
    }
  return TRUE;
}

/************************************************************************/
