/************************************************************************/
/*
Do dynamic programming to find policy for ODE simulation for the 
one link pendulum

We use a discount factor to make sure dynamic programming converges.
	      value = one_step_cost + DISCOUNT*lookup_value();
*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "useful.h"
#include "dynamics.h"
#include "dp.h"

/************************************************************************/

/* This sets how many actions eventually get explored in the grid search. */
#define MAX_N_ACTIONS_0 1001
/* This is how many actions we start out exploring in the grid search */
#define INITIAL_N_ACTIONS_0 10
/* How many sweeps we do */
#define N_SWEEPS 1000

#define MAX_TRAJ_LENGTH 1000

/************************************************************************/
/* Globals */

/* Current number of actions */
int n_actions[N_ACTION_DIMENSIONS] = { INITIAL_N_ACTIONS_0 }; 

// I am using 1D arrays for speed and simplicity
/* value functions: I double buffer these using v_k and v_next */
float v1[N_TOTAL_STATES];
float v2[N_TOTAL_STATES];
/* the current policy */
float u_array[N_TOTAL_STATES][N_ACTION_DIMENSIONS];
/* Where the policy takes us */
float next_state_array[N_TOTAL_STATES][N_STATE_DIMENSIONS];

/* pointers for swapping value functions */
float *v_k, *v_next;

float model_next_states[N_TOTAL_STATES][N_STATE_DIMENSIONS][N_ACTION_DIMENSIONS+1];

/* Keep track of total change */
float update2 = 0;

/* flag to indicate when we should examine a regular grid of actions */
int do_grid_search = 0;

float traj[MAX_TRAJ_LENGTH][N_STATE_DIMENSIONS];
int traj_length = 0;
float best_traj[MAX_TRAJ_LENGTH][N_STATE_DIMENSIONS];
int best_traj_length = 0;

/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
/* write all arrays to files */

int write_all_to_file()
{
  write_array_to_file( "v_k", v_k, N_TOTAL_STATES );
  write_array_to_file( "v_next", v_next, N_TOTAL_STATES );
  write_array_to_file( "next_state_array", &(next_state_array[0][0]),
		       N_TOTAL_STATES*N_STATE_DIMENSIONS );
  write_array_to_file( "u_array", &(u_array[0][0]),
		       N_TOTAL_STATES*N_ACTION_DIMENSIONS );
  return 0;
}

/************************************************************************/
/* initialization */

int init()
{
  int i, j;
	
  v_k = &(v1[0]);
  v_next = &(v2[0]);
		
  for( i = 0; i < N_TOTAL_STATES; i++ )
    {
      v_next[i] = 0;
      v_k[i] = 0;
      for( j = 0; j < N_ACTION_DIMENSIONS; j++ )
	u_array[i][j] = 0;
      for( j = 0; j < N_STATE_DIMENSIONS; j++ )
	next_state_array[i][j] = 0;
    }

  read_array_from_file( "model_next_states", 
			&(model_next_states[0][0][0]),
			N_TOTAL_STATES*(N_ACTION_DIMENSIONS+1)
			*N_STATE_DIMENSIONS );

  init_dynamics();
  init_dp();

  return 0;
}

/************************************************************************/
/* sweep function */

int do_sweep()
{
  int indices[N_STATE_DIMENSIONS];
  float best_action[N_ACTION_DIMENSIONS];
  float best_next_state[N_STATE_DIMENSIONS];
  int i, i1d;
  float v_best = INFINITE_COST;

  // increment n_actions 
  if ( n_actions[0] < MAX_N_ACTIONS_0 )
    (n_actions[0])++;

  // Keep track of changes to value function.
  update2 = 0;

  for( indices[0] = 0; indices[0] < n_states[0]; (indices[0])++ )
    {
      for( indices[1] = 0; indices[1] < n_states[1]; (indices[1])++ )
	{
	  i1d = value_function_index( indices );
	  if ( !lqr_covers_it( indices, NULL, &v_best, best_action,
			       best_next_state ) )
	    {
	      if ( do_grid_search )
		grid_search( indices, &v_best, best_action,
			     best_next_state, v_next, n_actions );
	      else
		{
		  v_best = v_k[i1d];
		  for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
		    best_action[i] = u_array[i1d][i];
		  for( i = 0; i < N_STATE_DIMENSIONS; i++ )
		    best_next_state[i] = next_state_array[i1d][i];
		  update_action( indices, &v_best, best_action,
				     best_next_state, v_next );
		  gradient_descent( indices, NULL,
				    &v_best, best_action,
				    best_next_state, v_next );
		  random_search( indices, &v_best, best_action,
				 best_next_state, v_next );
		}
	    }

	  // printf( "%d: v_best: %g; old: %lg\n", i1d, v_best, v_k[i1d] );
	  if ( v_best < COST_CUTOFF )
	    {
	      update2 += ( v_k[i1d] - v_best )*( v_k[i1d] - v_best );
	      v_k[i1d] = v_best;
	      // printf( "update2: %lg; %lg; %lg\n", update2, v_k[i1d], v_k[6] );
	      for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
		u_array[i1d][i] = best_action[i];
	      for( i = 0; i < N_STATE_DIMENSIONS; i++ )
		next_state_array[i1d][i] = best_next_state[i];
	    }
	  else
	    { /* Mark bad values */
	      v_k[i1d] = INFINITE_COST;
	      for( i = 0; i < N_ACTION_DIMENSIONS; i++ )
		u_array[i1d][i] = BAD_U;
	      for( i = 0; i < N_STATE_DIMENSIONS; i++ )
		next_state_array[i1d][i] = BAD_U;
	    }
	}
    }
  return 0;
}

/************************************************************************/
/************************************************************************/

float score_traj_using_policy()
{
  float score = 0;
  float action[N_ACTION_DIMENSIONS];
  float initial_state[N_STATE_DIMENSIONS];
  float value;
  int i, j;

  initial_state[0] = 0.0;
  initial_state[1] = 0;
  set_state( 0.0, initial_state );

  for( i = 0; ; i++ )
    {
      /* Stash trajectory */
      if ( i < MAX_TRAJ_LENGTH )
	{
	  for( j = 0; j < N_STATE_DIMENSIONS; j++ )
	    traj[i][j] = the_state[j];
	  traj_length = i+1;
	}

      if ( !lqr_covers_it( NULL, the_state, &value, action, NULL ) )
	lookup_action( the_state, u_array, v_k, action );
      score += one_step_cost( the_state, action );
      integrate_one_step( action, NULL );

      if ( the_time >= 9.999999 )
	{
	  break;
	}
    }
  return score;
}

/************************************************************************/
/************************************************************************/
/* main */

int main(int argc, char **argv)
{
  int i, j, k;
  float *temp;
  int index;
  float state[N_STATE_DIMENSIONS];
  int indices[N_STATE_DIMENSIONS];
  float score;
  float best_score = 1e30;
	
  init();
	
  /* Initialize value function */
  for( indices[0] = 0; indices[0] < n_states[0]; (indices[0])++ )
    for( indices[1] = 0; indices[1] < n_states[1]; (indices[1])++ )
      {
	index = value_function_index( indices );
	indices_to_state( indices, state );
	v_next[index] = lqr_value( state );
      }

  for( i = 0; i < N_SWEEPS; i++ )
    {
      do_sweep();
      printf( "Sweep: %d; Total update squared: %lg\n", i, update2 );
      
      /* Just to see how things are going, use the policy to generate
	 a trajectory and see how well it does */
      score = score_traj_using_policy();
      if ( score < best_score )
	{
	  best_score = score;
	  for( j = 0; j < traj_length; j++ )
	    for( k = 0; k < N_STATE_DIMENSIONS; k++ )
	      best_traj[j][k] = traj[j][k];
	  best_traj_length = traj_length;
	}
      printf( "Score using policy: %g (best: %g)\n", score, best_score );

      if ( i % 100 == 0 )
	{
	  write_all_to_file();
	}

      /* swap */
      temp = v_k;
      v_k = v_next;
      v_next = temp;
    }
  
  return 0;
}

/************************************************************************/

