/* 16-711 Demo program: Control the single axis robot.
 */
/* Author: Matthew McNaughton.
 */
/* This program is GPL'ed. Some GPL code taken from Sensoray demo app
 * s626demo.c. Insert GPL here.
 */

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>


extern "C" {
#include "s626drv.h"
#include "App626.h"
#include "s626mod.h"
#include "s626core.h"
#include "s626api.h"
#include "kalman.h"
}

/******************************* Global Data. ******************************/

/* The encoder input channel that the motor's encoder is hooked to:
 */
WORD encoder = CNTR_1B;

/* The DAC channel that is hooked to the amplifier.
 */
//int channel = 0;
int channel = 1; //markmp 3/28 - Matt changed the DAC channel

/* We are using the first board.
 */
WORD board = 0;

/* The adc input on the board that we're hooked into for the tachometer.
 * Adjust as necessary.
 */
WORD adc_channel = 0;

unsigned long		errFlags  = 0x0;		// error flags

/* Data file data structure */
FILE *vel_log;

/* Used to handle control-c */
int exit_flagged = 0;

int verbose = 0;

//markmp added constant voltage from command line
float g_constant_voltage = 0;

//markmp - added for special drive modes
int g_mode = 0;

//take command line values for columb and viscous friction
float g_const_Cc = -1;
float g_const_Cv = -1;
float g_const_Cg = -1;

/************************** End Global Data. **************************/

void
catch_int_c(int signum) {
  exit_flagged = 1;
}

void
stop_motor_and_exit() {
  S626_WriteDAC (board, channel, 0); /* Set zero torque. */
  if( vel_log ) fflush(vel_log);
  exit(0);
}

VOID ErrorFunction1 (DWORD ErrFlags)
{
  printf ("Got an error on board %d 0x%lx\n", board, ErrFlags);
}

int
OpenBoard() {

  S626_OpenBoard( board, 0, /*InterruptAppISR*/ 0, 1 );
  S626_InterruptEnable( board, FALSE );
  S626_SetErrCallback( board, ErrorFunction1 );
  errFlags = S626_GetErrors (board);
  if(verbose || errFlags ) 
    printf("Board(%i) -  ErrFlags = 0x%lx \n", board, errFlags );

  if (errFlags!=0x0) 
  {
	  switch (errFlags)
	  {
		  case ERR_OPEN:
			  printf("\t\t Can't open driver.\n"); break;
		  case ERR_CARDREG:
			  printf("\t\t Can't attach driver to board.\n"); break;
		  case ERR_ALLOC_MEMORY:
			  printf("\t\t Memory allocation error.\n"); break;
		  case ERR_LOCK_BUFFER:
			  printf("\t\t Error locking DMA buffer.\n"); break;
		  case ERR_THREAD:
			  printf("\t\t Error starting a thread.\n"); break;
		  case ERR_INTERRUPT:
			  printf("\t\t Error enabling interrupt.\n"); break;
		  case ERR_LOST_IRQ:
			  printf("\t\t Missed interrupt.\n"); break;
		  case ERR_INIT:
			  printf("\t\t Board object not instantiated.\n"); break;
		  case ERR_SUBIDS:
			  printf("\t\t PCI SubDevice/SubVendor ID mismatch.\n"); break;
		  case ERR_CFGDUMP:
			  printf("\t\t PCI configuration dump failed.\n"); break;
		  default:
			  printf("\t\t other Unknown errors.\n"); break;
	  }
	  printf("Fix board open/installation error, and then try again. \n\n");
	  return -1;
  }
  /* Success. */
  return 0;
}


/* Set up the encoder.
 */
void CreateEncoderCounter( HBD hbd, WORD Counter )
{
	// Set operating mode for the given Counter.
	S626_CounterModeSet( hbd, Counter,
		( LOADSRC_INDX << BF_LOADSRC ) |    // Preload upon index.
		( INDXSRC_SOFT << BF_INDXSRC ) |    // Disable hardware index.
		( CLKSRC_COUNTER << BF_CLKSRC ) |   // Operating mode is Counter.
		( CLKPOL_POS  << BF_CLKPOL ) |      // Active high clock.
		//( CNTDIR_UP  << BF_CLKPOL ) |       // Count direction is Down.
		( CLKMULT_1X   << BF_CLKMULT ) |    // Clock multiplier is 1x.
		( CLKENAB_INDEX << BF_CLKENAB ) );  // Counting is initially disabled.

	// Set counter core and preload value to be mid of 2^24 (since all counters are 24-bit), so as to make test easy.
	S626_CounterPreload( hbd, Counter, 8388608 );	// 0x800000 = 2^24/2 = 8388608
	S626_CounterSoftIndex( hbd, Counter );			// Generate a index signal by software, 
													// so that the preload value got loaded when creating EncoderCounter.

	// Enable latching of accumulated counts on demand.
	S626_CounterLatchSourceSet( hbd, Counter, LATCHSRC_AB_READ );

	// Enable the counter.
	S626_CounterEnableSet( hbd, Counter, CLKENAB_ALWAYS );
}



/**
 * Only works on ranges of a few dozens of seconds time difference.
 */
long long
usec_diff( struct timeval *tv1, struct timeval *tv2 ) {
  long long result = tv1->tv_usec - tv2->tv_usec;
  if( result < 0 ) {
    result += 1000000;
    result += 1000000 * (tv1->tv_sec - tv2->tv_sec - 1);
  } else {
    result += 1000000 * (tv1->tv_sec - tv2->tv_sec);
  }

  return result;
}

void
busy_wait_until( struct timeval basetime, long long until ) {
  struct timeval tv;
  long long now;
  do {
    gettimeofday( &tv, 0 );
    now = usec_diff( &tv, &basetime );
    if( exit_flagged ) stop_motor_and_exit();
  } while( now < until );
}

/* There should only be one actively used instance of these in the system.
 */
typedef struct {
  /* used to setup the ADC.
   */
  BYTE				poll_list[16];

  /* holds the values read by the ADC.
   */
  WORD				data[16];
} adc_t;

/*
 * chanmask is a mask.
 * Assumes 10 volt range.
 */
void
adc_set_channels( adc_t *adc, unsigned int chanmask ) {
  int i; int nchan = 0;
  for( i = 0; i < 16; ++i ) {
    if( chanmask & (1 << i) ) {
      adc->poll_list[ nchan++ ] = (ADC_RANGE_10V << 4) | i;
    }
  }
  if( nchan > 0 ) {
    adc->poll_list[ nchan-1 ] |= ADC_EOPL;
  } else {
    fprintf(stderr, "adc_set_channels needs at least one channel.\n");
    exit(1);
  }

  S626_ResetADC (board, &adc->poll_list[0]);	//set up the poll list
}

void
sample_loop( float target, float gain_k, float gain_b, float maxvolt) {
  adc_t adc;
  int cnt = 0;
  int last_encoder = 0;
  struct timeval last_encoder_time;
  struct timeval first_encoder_time;
  struct timeval encoder_time;
  int lastPos = 0;
  float voltage = 0;
  double kalVel;

  const int hz1 = 10;
  long long next_time_step = 0;
  
  /* Reset Counter
   */
  S626_CounterCapFlagsReset( board, encoder );

  /* Configure counter xA/xB as a encoder counter.
   */
  CreateEncoderCounter( board, encoder );

  adc_set_channels( &adc, (1 << adc_channel) );

  gettimeofday( &last_encoder_time, 0 );
  first_encoder_time = last_encoder_time;

  //int first = 1;

  kalmanInit();

  while(1) {

    if( exit_flagged ) stop_motor_and_exit();

    /* Read the current position.
     */
    /* 8388608 is 2^23, in the middle of the range.
     */
    S626_ReadADC (board, &adc.data[0]);
    int current_encoder = S626_CounterReadLatch(board,encoder)-8388608;
    gettimeofday( &encoder_time, 0 );

    long long usecs = usec_diff( &encoder_time, &last_encoder_time );
    long long totalusecs = usec_diff( &encoder_time, &first_encoder_time );
    double tdouble = (double)totalusecs / (double)1e6;
    double diffDouble = (double)usecs / (double)1e6;

    /* Make the data signed.
     */
    int sign_adc = (short)adc.data[0];

    int rawpos = current_encoder;
    int rawvel = sign_adc;

    /* DO YOUR STUFF HERE. */

    //float voltage = 0;
    voltage = 0;
    voltage += gain_k * (target - rawpos);
    voltage += gain_b * (- rawvel);

    //voltage = 0.175;

    //if we have a constant voltage from the command line use that
    if (g_constant_voltage > 0.0f)
    {
      voltage = g_constant_voltage;
    }

    if (g_mode == 1) //compensate for friction and gravity mode
    {
      //number of ticks per radian (1000 per revolution)
      float positionInRadians = rawpos / 159.15494f; 
      //float veloInRadCenSec = rawvel * ((float)(4.5e-5));
      kalVel = kalmanUpdate(diffDouble, (double)rawpos/159.15494, (double)lastPos/159.15494, voltage);
      double veloInRadCenSec = kalVel;

      float velocitySign = 1;
      if (veloInRadCenSec < 0)
	velocitySign = -1;
      
      //model params from hw4 - may need to remodel for hw5 since matt changed amp
      //double Cg = 0.72004226501331;
      double Cg = 0.5;
      //double Cc = 0.22329504549906; //0.12329504549906;
      double Cc = 0.2;
      //double Cv = 0.01339875306335;
      double Cv = 0.001;
      //const double Ci = 0.01944485646583;
      double frictionTorque;

      if (g_const_Cc > 0) Cc = g_const_Cc;
      if (g_const_Cv > 0) Cv = g_const_Cv;
      if (g_const_Cg > 0) Cg = g_const_Cg;
       
      
      //at point zero - columb is moving pendulm
      if (positionInRadians < .05 || veloInRadCenSec < .00005)
      {
	voltage = 0;
      }
      else
      {
	frictionTorque = -Cg*sin(positionInRadians)
	  - Cc*velocitySign
	  - Cv*veloInRadCenSec;
	voltage = -1*frictionTorque;
      }
    }
    
    lastPos = rawpos;

    /* FINISH YOUR STUFF */

    /* Cap voltage to between +/- maxvolt.
     */
    if( voltage < -maxvolt ) voltage = -maxvolt;
    else if( voltage > maxvolt ) voltage = maxvolt;

    /* The DAC takes corresponding input in the rage -8190 to 8191.
     */
    int dacvalue = (int)rint((voltage / 10.0) * 8191.0);
    if( dacvalue < -8190 ) dacvalue = -8190;
    else if( dacvalue > 8190 ) dacvalue = 8190;

    S626_WriteDAC (board, channel, dacvalue);


    if( verbose && totalusecs > cnt * 100000 ) {
      cnt++;
      printf("enc=%d, vel=%d, sadc=%d, usecs=%lld, volt=%f=%d\n",
	     current_encoder, rawvel, sign_adc, usecs, voltage, dacvalue);
    }

    last_encoder = current_encoder;
    last_encoder_time = encoder_time;

    if( vel_log != 0 ) {
      fprintf(vel_log, "%f %f %f %f %f %.3f\n", 
	      tdouble, (float) rawpos, (float) rawvel, kalVel, voltage );
    }

    next_time_step += hz1 * 1000;
    busy_wait_until( first_encoder_time, next_time_step );
  }
  kalmanDestroy();
}



int
main( int argc, char *argv[] ) {
  int optchar;

  float gain_k = 0.01, gain_b = 0.001, maxvolt = 1;
  float target_param = 0;

  /* Read servo gains from command line.
   */

  while( (optchar = getopt(argc, argv, "a:k:b:t:l:v:c:m:g:")) != -1 ) {
    switch( optchar ) {
      //case 'v':
      //verbose = 1;
      //break;
    case 'v':
      if( sscanf(optarg, "%f", &g_const_Cv) != 1 ) {
	fprintf(stderr, "-v option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'c':
      if( sscanf(optarg, "%f", &g_const_Cc) != 1 ) {
	fprintf(stderr, "-c option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'g':
      if( sscanf(optarg, "%f", &g_const_Cg) != 1 ) {
	fprintf(stderr, "-g option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'k':
      if( sscanf(optarg, "%f", &gain_k) != 1 ) {
	fprintf(stderr, "-k option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'b':
      if( sscanf(optarg, "%f", &gain_b) != 1 ) {
	fprintf(stderr, "-b option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 't':
      if( sscanf(optarg, "%f", &target_param ) != 1 ) {
	fprintf(stderr, "-t option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'a':
      if( sscanf(optarg, "%f", &g_constant_voltage ) != 1 ) {
	fprintf(stderr, "-a option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
    case 'm':
      if( sscanf(optarg, "%d", &g_mode ) != 1 ) {
	fprintf(stderr, "-m option %s is not an int.\n", optarg);
	exit(1);
      }
      break;
      /*
    case 'm':
      if( sscanf(optarg, "%f", &maxvolt ) != 1 ) {
	fprintf(stderr, "-m option %s is not a float.\n", optarg);
	exit(1);
      }
      break;
      */

    case 'l':
      /* log file
       */
      vel_log = fopen( optarg, "w" );
      if( vel_log == 0 ) {
	fprintf(stderr, "Could not open %s for writing: %s\n",
		optarg, strerror(errno));
      }
      break;
      
    case '?':
      fprintf(stderr, "bad option %c. rtfs.\n", optopt);
      exit(1);
    }
  }

  if( OpenBoard() != 0 ) {
    exit(1);
  }

  /* When the program is terminated, we'll make the motor stop.
   */
  signal( SIGINT, catch_int_c );


  sample_loop( target_param, gain_k, gain_b, maxvolt);
  stop_motor_and_exit();
}

